#include <windows.h>
#include <stdio.h>
#include "kmp_com.h"
#include "xstdlib.h"
#include "mxdrvg/mxdrv.h"
#include "lzx/lzx042.h"
#include "mdc/mdc.h"

#define KPI_VERSION 11
#define KPI_VERSIONS "r11"

#define MAX_TITLE 512
#define MAX_PDXNAME 512

struct KMP_CTX_TAG
{
	HMXDRVG *hMxdrv;
	unsigned srate;
	unsigned nch;
	unsigned bits;
	unsigned length;
	unsigned loop;
	BOOL bEnableLpf;
};

static const char *regsubkey1 = "Software\\Kobarin's Soft\\KbMedia Player\\Path";
static const char *regsubkey2 = "Software\\Kobarin's Soft\\KbMIDI Player\\SearchPath";
static const char *sectionname = "mxdrvg";
static char iniFilePath[MAX_PATH] = "";

static UINT GetPrivateProfileIntEx(LPCTSTR lpAppName, LPCTSTR lpKeyName, INT nDefault, LPCTSTR lpFileName)
{
	UINT ret;
	TCHAR szBuffer[32];
	ret = GetPrivateProfileInt(lpAppName, lpKeyName, nDefault, lpFileName);
	wsprintf(szBuffer, "%d", ret);
	WritePrivateProfileString(lpAppName, lpKeyName, szBuffer, lpFileName);
	return ret;
}

static BOOL GetPrivateProfileBoolEx(LPCTSTR lpAppName, LPCTSTR lpKeyName, BOOL bDefault, LPCTSTR lpFileName)
{
	BOOL ret;
	TCHAR szBuffer[32];
	if (GetPrivateProfileString(lpAppName, lpKeyName, bDefault ? "True" : "False", szBuffer, 32, lpFileName))
	{
		/* 擪mF */
		ret = szBuffer[0] == 'T' || szBuffer[0] == 't';
	}
	else
		ret = bDefault;
	WritePrivateProfileString(lpAppName, lpKeyName, ret ? "True" : "False", lpFileName);
	return ret;
}

static int RegGetString(const char *lpszSubKey, const char *lpszName, char *lpszBuffer, unsigned uBufferSize)
{
	int ret = 0;
	HKEY hkey;
	DWORD dwt, dws;
	if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_CURRENT_USER, lpszSubKey, 0, KEY_QUERY_VALUE, &hkey)) return ret;
	dws = uBufferSize;
	if (ERROR_SUCCESS == RegQueryValueEx(hkey, lpszName,NULL,&dwt,(LPBYTE)lpszBuffer, &dws) && dwt == REG_SZ)
	{
		ret = dws;
	}
	RegCloseKey(hkey);
	return ret;
}

static void Start(KMP_CTX *pctx)
{
}

static void Setup(KMP_CTX *pctx, unsigned srate)
{
	pctx->hMxdrv = 0;
	pctx->nch = 2;
	pctx->bits = 16;
	pctx->loop = GetPrivateProfileIntEx(sectionname, "Loop", 2, iniFilePath);
	if (11025 <= srate && srate <= 96000)
		pctx->srate = srate;
	else
		pctx->srate = GetPrivateProfileIntEx(sectionname, "SampleRate", 44100, iniFilePath);
	pctx->bEnableLpf = GetPrivateProfileBoolEx(sectionname, "EnableLpf", TRUE, iniFilePath);
}

__inline void SetWordBE(unsigned char *p, unsigned v)
{
	p[0] = (unsigned char)(v >> 8);
	p[1] = (unsigned char)(v);
}

__inline unsigned GetWordBE(const unsigned char *p)
{
	return ((unsigned)p[0] << 8) + p[1];
}

int GetDwordBE(char *p)
{
	int ret;
	ret  = ((int)(unsigned char)p[0]) << 0x18;
	ret |= ((int)(unsigned char)p[1]) << 0x10;
	ret |= ((int)(unsigned char)p[2]) << 0x08;
	ret |= ((int)(unsigned char)p[3]) << 0x00;
	return ret;
}

static void *load_file(const char *path, unsigned *psize)
{
	XFILE fp = 0;
	void *ret = 0;
	void *buf = 0;
	do
	{
		unsigned size;
		fp = XOPENRB(path);
		if (!fp) break;
		size = XGETSIZE(fp);
		if (!size) break;
		buf = XMALLOC(size + 8);
		if (!buf) break;
		if (size != XREAD(fp, buf, size)) break;
		if (psize) *psize = size;
		ret = buf; buf = 0;
	} while (0);
	if (buf) XFREE(buf);
	if (fp) XCLOSE(fp);
	return ret;
}

static void *load_pdxfile_sub(char *path, unsigned *psize)
{
	unsigned char *ret = 0, *pdxbuf = 0, *pdxfile = 0;
	do
	{
		unsigned lzxsize, pdxsize, pdxbufsize, pdxnamelen, pdxbodyptr, pdxbodysize;
		pdxfile = load_file(path, &pdxsize);
		if (!pdxfile) break;
		lzxsize = lzx042check(pdxfile);
		pdxnamelen = XSTRLEN(path) + 1;
		pdxbodyptr = 8 + pdxnamelen;
		pdxbodysize = (lzxsize ? lzxsize : pdxsize);
		pdxbufsize = pdxbodyptr + pdxbodysize;
		/* pdxbufsize = (pdxbufsize + 1024) & ~ 1023; */
		pdxbuf = XMALLOC(pdxbufsize * 2 + 8 + 1);
		if (!pdxbuf) break;
		if (((unsigned)(pdxbuf + pdxbodyptr)) & 1)
		{
			pdxbodyptr++;
			pdxbufsize++;
		}
		XMEMSET(pdxbuf, 0, pdxbodyptr);
		XMEMCPY(pdxbuf + 8, path, pdxnamelen);
		if (lzxsize)
			lzx042decode(pdxbuf + pdxbodyptr, lzxsize, pdxfile, pdxsize);
		else
			XMEMCPY(pdxbuf + pdxbodyptr, pdxfile, pdxsize);
		SetWordBE(pdxbuf + 0, 0);
		SetWordBE(pdxbuf + 2, 0);
		SetWordBE(pdxbuf + 4, pdxbodyptr);
		SetWordBE(pdxbuf + 6, pdxnamelen);
		*psize = pdxbodysize;
		ret = pdxbuf; pdxbuf = 0;
	} while (0);
	if (pdxfile) XFREE(pdxfile);
	if (pdxbuf) XFREE(pdxbuf);
	return ret;
}

static void *load_pdxfile(char *path, char *fname, unsigned *psize)
{
	char *pathbuf;
	void *ret = 0;
	pathbuf = XMALLOC(XSTRLEN(path) + XSTRLEN(fname) + 4 + 1);
	if (pathbuf)
	{
		XSTRCPYN(pathbuf, path, ~0);
		XSTRCPYN(pathbuf + XSTRLEN(pathbuf), fname, ~0);
		ret = load_pdxfile_sub(pathbuf, psize);
		if (!ret)
		{
			XSTRCPYN(pathbuf + XSTRLEN(pathbuf), ".pdx", ~0);
			ret = load_pdxfile_sub(pathbuf, psize);
		}
		XFREE(pathbuf);
	}
	return ret;
}

static int Load(KMP_CTX *pctx, const char *mdxfilename)
{
	unsigned char *mdxfile = 0;
	unsigned char *mdxbuf = 0;
	unsigned char *pdxbuf = 0;

	do
	{
		char mdxpath[_MAX_PATH+32];
		char headtitle[MAX_TITLE+32];
		char headpdxname[MAX_PDXNAME+32];

		unsigned mdxsize, mdxbufsize, lzxsize, pdxbufsize;
		unsigned mdxtitlelen, mdxbodyptr, mdxbodysize;
		unsigned char *p;
		unsigned char *p2;
		unsigned char c;

		XDIVPATH(mdxpath, mdxfilename, sizeof(mdxpath));
		mdxfile = load_file(mdxfilename, &mdxsize);
		if (!mdxfile || mdxsize <= 3 + 1 + 20) break;
		/* MDC to MDX */
		if (GetDwordBE(mdxfile) == 0x4d44431a)
		{
			mdxfile = mdc2mdx(mdxfile, &mdxsize);
			if (!mdxfile || mdxsize <= 3 + 1 + 20) break;
		}
		/* SKIP TITLE */
		p = mdxfile;
		while (mdxsize)
		{
			c = *(p++);
			if (--mdxsize == 0) break;
			if (c == 0x0d) break;
			if (c == 0x0a) break;
			if (c < 0x20 && c != 0x1b)
			{
				mdxsize = 0;
				break;
			}
		}
		if (mdxsize == 0) break;
		p[-1] = 0x00;
		XSTRCPYN(headtitle, mdxfile, MAX_TITLE);
		if ((p - mdxfile) & 1)
		{
			*(p++) = 0x00;
			if (--mdxsize == 0) break;
		}
		p2 = p;
		if (c != 0x0d)
		{
			while (mdxsize && *p2 != 0x0d) p2++, mdxsize--;
			if (mdxsize == 0) break;
			p2++;
			if (--mdxsize == 0) break;
		}
		while (mdxsize && *p2 != 0x1a) p2++, mdxsize--;
		if (mdxsize == 0) break;
		p2++;
		if (--mdxsize == 0) break;
		/* LOAD PDX */
		XSTRCPYN(headpdxname, p2, MAX_PDXNAME);
		if (*headpdxname) 
		{
			pdxbuf = load_pdxfile(mdxpath, headpdxname, &pdxbufsize);
			if (!pdxbuf)
			{
				char pdxsearchpath[_MAX_PATH];
				if ((RegGetString(regsubkey1, "PDX", pdxsearchpath, sizeof(pdxsearchpath) - 1) || RegGetString(regsubkey2, "PDX", pdxsearchpath, sizeof(pdxsearchpath) - 1)) && pdxsearchpath[0])
				{
					int pos;
					pos = XSTRLEN(pdxsearchpath) - 1;
					if (pdxsearchpath[pos] != '\\' && pdxsearchpath[pos] != '/')
					{
						pdxsearchpath[pos++] = '\\';
						pdxsearchpath[pos] = '\0';
					}
					pdxbuf = load_pdxfile(pdxsearchpath, headpdxname, &pdxbufsize);
				}
			}
		}
		if (!pdxbuf) pdxbufsize = 0;
		/* SKIP PDX NAME */
		while (mdxsize && *p2) p2++, mdxsize--;
		if (mdxsize == 0) break;
		p2++;
		if (--mdxsize == 0) break;

		lzxsize = lzx042check(p2);
		mdxtitlelen = XSTRLEN(headtitle) + 1;
		mdxbodyptr = 8 + mdxtitlelen;
		mdxbodysize = (lzxsize ? lzxsize : mdxsize);
		mdxbufsize = mdxbodyptr + mdxbodysize;
		/* mdxbufsize = (mdxbufsize + 1024) & ~ 1023; */
		mdxbuf = XMALLOC(mdxbufsize * 2 + 8 + 1);
		if (!mdxbuf) break;
		if (((unsigned)(mdxbuf + mdxbodyptr)) & 1)
		{
			mdxbodyptr++;
			mdxbufsize++;
		}
		XMEMSET(mdxbuf, 0, mdxbodyptr);
		XMEMCPY(mdxbuf + 8, headtitle, mdxtitlelen);
		if (lzxsize)
			lzx042decode(mdxbuf + mdxbodyptr, lzxsize, p2, mdxsize);
		else
			XMEMCPY(mdxbuf + mdxbodyptr, p2, mdxsize);
		SetWordBE(mdxbuf + 0, 0);
		SetWordBE(mdxbuf + 2, pdxbuf ? 0 : 0xffff);
		SetWordBE(mdxbuf + 4, mdxbodyptr);
		SetWordBE(mdxbuf + 6, 8);
		XFREE(mdxfile);	mdxfile = 0;
		/*
			Đg(Hz)
			(ms) gp(0)
			PCMϊobt@(ms)
			x(ms) gp
			MDXobt@TCY(byte)
			PDXobt@TCY(byte)
		*/
		pctx->hMxdrv = MXDRVG_Start(pctx->srate, pctx->bEnableLpf, mdxbufsize * 2, pdxbufsize * 2);
		if (!pctx->hMxdrv) break;

		if (1/*pdxbuf*/)
		{
			unsigned expcm;
			expcm = GetWordBE(&mdxbuf[mdxbodyptr + 2]);
			if (expcm < mdxbufsize)
				expcm = mdxbuf[mdxbodyptr + expcm] == 0xe8;
			else
				expcm = 0;
			p2 = (unsigned char *)MXDRVG_GetWork(pctx->hMxdrv, MXDRV_WORK_PCM8);
			*p2 = expcm ? 1 : 0;
		}
		/* MXDRVG_Stop(pctx->hMxdrv); */
		pctx->length = MXDRVG_MeasurePlayTime(pctx->hMxdrv, mdxbuf, mdxbufsize, pdxbuf, pdxbufsize, pctx->loop, 1);
		MXDRVG_Play(pctx->hMxdrv, mdxbuf, mdxbufsize, pdxbuf, pdxbufsize);
		if (mdxbuf) XFREE(mdxbuf);
		if (pdxbuf) XFREE(pdxbuf);
		return 1;
	} while (0);
	if (mdxbuf) XFREE(mdxbuf);
	if (mdxfile) XFREE(mdxfile);
	if (pdxbuf) XFREE(pdxbuf);
	return 0;
}

void Close(KMP_CTX *pctx)
{
	if (pctx)
	{
		if (pctx->hMxdrv)
		{
			MXDRVG_End(pctx->hMxdrv);
			pctx->hMxdrv = 0;
		}
		XFREE(pctx);
	}
}

KMP_CTX *Open(const char *file, unsigned size, unsigned srate)
{
	KMP_CTX *pctx = 0;
	do
	{
		pctx = XMALLOC(sizeof(KMP_CTX));
		if (!pctx) break;
		Setup(pctx, srate);

		if (size)
		{
			/* ̓ǂݍ݂̓T|[gȂ */
			break;
		}
		else
		{
			if (!Load(pctx, file)) break;
		}
		Start(pctx);
		return pctx;
	} while (0);
	if (pctx) Close(pctx);
	pctx = 0;
	return pctx;
}

unsigned Write(KMP_CTX *pctx, void *buf, unsigned smp)
{
	MXDRVG_GetPCM(pctx->hMxdrv, buf, smp >> 2);
	return smp;
}

unsigned WriteSkip(KMP_CTX *pctx, unsigned smp)
{
	BYTE tmpbuf[512 * 4];
	unsigned r = smp;
	while (r > sizeof(tmpbuf))
	{
		MXDRVG_GetPCM(pctx->hMxdrv, tmpbuf, sizeof(tmpbuf) >> 2);
		r -=  sizeof(tmpbuf);
	}
	if (r)
	{
		MXDRVG_GetPCM(pctx->hMxdrv, tmpbuf, r >> 2);
	}
	return smp;
}

unsigned SetPosition(KMP_CTX *pctx, unsigned ms)
{
	MXDRVG_PlayAt(pctx->hMxdrv, ms, pctx->loop, 0);
	return ms;
}

int GetLoopFlag(KMP_CTX *pctx)
{
	return 1;
}

int GetSeekableFlag(KMP_CTX *pctx)
{
	return 1;
}

int GetMultisongFlag(KMP_CTX *pctx)
{
	return 0;
}

unsigned GetFrequency(KMP_CTX *pctx)
{
	return pctx->srate;
}
unsigned GetChannels(KMP_CTX *pctx)
{
	return pctx->nch;
}
unsigned GetSamplebits(KMP_CTX *pctx)
{
	return pctx->bits;
}
unsigned GetLength(KMP_CTX *pctx)
{
	return pctx->length;
}
unsigned GetUnitSamples(KMP_CTX *pctx)
{
	return (pctx->srate * 60) / 1000;	/* 60msec */
}

void Init(void)
{
}
void Deinit(void)
{
}
void XCRTINIT(void)
{
	unsigned pos;
	static const char *iniext = ".ini";
	pos = XDELEXT(iniFilePath, XGETMODULEPATH(), sizeof(iniFilePath));
	XSTRCPYN(iniFilePath + pos, iniext, sizeof(iniFilePath) - pos);
}
void XCRTTERM(void)
{
}

const char **GetPluginInfo(void)
{
	static char flags[4] = { KPI_VERSION, 1, 0, 0 };
	static const char *infos[] = {
		flags,
		"mxdrvg.kpi v2.00b" KPI_VERSIONS " by Mamiya",
		"X68k MXDRV music driver version 2.06+17 Rel.X5-S Copyright (c)1988-92 milk.,K.MAEKAWA, Missy.M, Yatsube / MXDRVg V2.00b Copyright (C) 2000-2002 GORRY / X68Sound.dll 20020615 by m_puusan",
		".mdx",".mdc",0
	};
	if (!GetPrivateProfileBoolEx(sectionname, "EnableMDC", TRUE, iniFilePath)) infos[4] = 0;
	flags[1] = GetPrivateProfileBoolEx(sectionname, "EnableReentrant", TRUE, iniFilePath);
	return infos;
}
