/*
  [Win32ˑ]
    DLL̏R[hƃAP[^łB
*/
#define XMEMSAFE 0

#include "win32l.h"
#if XMEMSAFE
#undef NOUSER
#undef NOMB
#endif
#undef NONLS
#include <windows.h>
#include "xstdlib.h"

#if defined(_MSC_VER) && !defined(_DEBUG)
#pragma comment(linker,"/OPT:NOWIN98")
#endif

static CRITICAL_SECTION globallock;
void XGLOBALLOCK(void)
{
	EnterCriticalSection(&globallock);
}
void XGLOBALUNLOCK(void)
{
	LeaveCriticalSection(&globallock);
}

static char dllpath[MAX_PATH] = "";
char *XGETMODULEPATH(void)
{
	return dllpath;
}

static HINSTANCE hDLLInstance = 0;
void *XGETMODULEINSTANCE(void)
{
	return hDLLInstance;
}

#if XMEMSAFE
static char *MemSafeId1 = "MemSafe\0\xF1\xE2\xD3\xC4";
static char *MemSafeId2 = "MemSafe\0\x1F\x2E\x3D\x4C";
static LONG LeakCheckCounter = 0;
enum {
	MSE_LEAK,
	MSE_NPFREE,
	MSE_IPFREE,
	MSE_BPFREE,
	MSE_NPREALLOC,
	MSE_IPREALLOC,
	MSE_BPREALLOC
};
static void MemSafeError(int code)
{
	static const char *(errmes[]) = {
		"MemoryLeak",
		"NULL pointer [free]",
		"Invalid pointer [free]",
		"Broken pointer [free]",
		"NULL pointer [realloc]",
		"Invalid pointer [realloc]",
		"Broken pointer [realloc]",
	};
	MessageBoxA(0,errmes[code],"MemSafe - Error report",MB_OK);
}
static void SetDwordLE(unsigned char *p, unsigned v)
{
	p[0] = (v >> (8 * 0)) & 0xFF;
	p[1] = (v >> (8 * 1)) & 0xFF;
	p[2] = (v >> (8 * 2)) & 0xFF;
	p[3] = (v >> (8 * 3)) & 0xFF;
}
static unsigned GetDwordLE(const unsigned char *p)
{
	return p[0] | (p[1] << 8) | (p[2] << 16) | (p[3] << 24);
}
#endif

static HANDLE hheap = 0;
static void XHEAPINIT(void)
{
	if (!hheap) hheap = HeapCreate(0, 0, 0);
}

static void XHEAPTERM(void)
{
#if XMEMSAFE
	if (LeakCheckCounter) MemSafeError(MSE_LEAK);
#endif
	if (hheap)
	{
		HeapDestroy(hheap);
		hheap = 0;
	}
}

void *XMALLOC(unsigned l)
{
#if XMEMSAFE
	unsigned char *head, *tail;
	head = HeapAlloc(hheap ? hheap : GetProcessHeap(), 0, l + 32);
	if (!head) return 0;
	InterlockedIncrement(&LeakCheckCounter);
	/* write id */
	tail = head + 16 + l;
	XMEMCPY(head, MemSafeId1, 12);
	SetDwordLE(head + 12, l);
	XMEMCPY(tail, MemSafeId2, 12);
	SetDwordLE(tail + 12, l);
	return head + 16;
#else
	return HeapAlloc(hheap ? hheap : GetProcessHeap(), 0, l);
#endif
}

void *XREALLOC(void *p, unsigned l)
{
#if XMEMSAFE
	unsigned char *head, *tail;
	unsigned oldl;
	if (!p)
	{
		MemSafeError(MSE_NPREALLOC);
		return 0;
	}
	/* check id */
	head = ((unsigned char *)p) - 16;
	tail = head + 16 + (oldl = GetDwordLE(head + 12));
	if (XMEMCMP(head, MemSafeId1, 12))
	{
		MemSafeError(MSE_IPREALLOC);
		return 0;
	}
	if (XMEMCMP(tail, MemSafeId2, 12) || XMEMCMP(head + 12, tail + 12, 4))
	{
		MemSafeError(MSE_BPREALLOC);
		return 0;
	}
	/* clear id */
	XMEMSET(head, 0, 16);
	XMEMSET(tail, 0, 16);
	p = HeapReAlloc(hheap ? hheap : GetProcessHeap(), 0, head, l + 32);
	if (p)
	{
		head = p;
		tail = head + 16 + l;
	}
	else
	{
		l = oldl;
	}
	/* rewrite id */
	XMEMCPY(head, MemSafeId1,12);
	SetDwordLE(head + 12, l);
	XMEMCPY(tail, MemSafeId2,12);
	SetDwordLE(tail + 12, l);
	return p ? head + 16 : 0;
#else
	return HeapReAlloc(hheap ? hheap : GetProcessHeap(), 0, p, l);
#endif
}

void XFREE(void *p)
{
#if XMEMSAFE
	unsigned char *head, *tail;
	if (!p)
	{
		MemSafeError(MSE_NPFREE);
		return;
	}
	/* check id */
	head = ((unsigned char *)p) - 16;
	tail = head + 16 + GetDwordLE(head + 12);
	if (XMEMCMP(head, MemSafeId1, 12))
	{
		MemSafeError(MSE_IPFREE);
		return;
	}
	if (XMEMCMP(tail, MemSafeId2, 12) || XMEMCMP(head + 12, tail + 12, 4))
	{
		MemSafeError(MSE_BPFREE);
		return;
	}
	/* clear id */
	XMEMSET(head, 0, 16);
	XMEMSET(tail, 0, 16);
	InterlockedDecrement(&LeakCheckCounter);
	HeapFree(hheap ? hheap : GetProcessHeap(), 0, head);
#else
	HeapFree(hheap ? hheap : GetProcessHeap(), 0, p);
#endif
}


BOOL WINAPI DllMain(
	HINSTANCE hinstDLL,
	DWORD fdwReason,
	LPVOID lpvReserved
)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
	{
		InitializeCriticalSection(&globallock);
		DisableThreadLibraryCalls(hinstDLL);
		XHEAPINIT();
		hDLLInstance = hinstDLL;
		GetModuleFileName(hinstDLL, dllpath, sizeof(dllpath));
		XCRTINIT();
	}
	else if (fdwReason == DLL_PROCESS_DETACH)
	{
		XCRTTERM();
		XHEAPTERM();
		DeleteCriticalSection(&globallock);
	}
	return TRUE;
}
